/*
 * Test filesystem listing concurrency limitations.
 *
 * Copyright (C) 2019, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <ext2fs/ext2fs.h>

#include "file.h"
#include "format.h"
#include "image.h"
#include "path.h"
#include "utils.h"



/* Show directory entries when iterating. */

static int image_list_dir_proc(struct ext2_dir_entry *dirent, int offset,
                               int blocksize, char *buf, void *priv_data)
{
    struct utils_list_dir_data *data = (struct utils_list_dir_data *) priv_data;
    ext2_filsys fs = (ext2_filsys) data->fs;
    struct ext2_inode inode;
    ext2_file_t file;
    __u64 size;
    char *to_remove[] = {"file0061", "file0062", "file0063", "file0125", "file0126", "file0127", 0};
    char **f;

    /* Call the formatting utility function. */

    utils_list_dir_proc(dirent, offset, blocksize, buf, priv_data);

    /* Remove files at an awkward point in time. */

    if (!strncmp(dirent->name, "file0060", ext2fs_dirent_name_len(dirent)))
    {
        for (f = to_remove; *f; f++)
            if (ext2fs_unlink(fs, EXT2_ROOT_INO, *f, 0, 0))
                printf("%s not removed!\n", *f);
    }

    return 0;
}



int main(int argc, char *argv[])
{
    int flags = EXT2_FLAG_RW;
    char *fsname;
    ext2_filsys fs = NULL;
    errcode_t retval;
    int exitcode = 0;
    char basename[32];
    ext2_ino_t ino;
    ext2_file_t file;
    int i;
    struct utils_list_dir_data list_dir_data;

    if (argc < 2)
    {
        printf("Usage: %s <device or image file>\n", argv[0]);
        return 1;
    }

    fsname = argv[1];

    /* Open the filesystem image using the POSIX file access mechanism. */

    retval = ext2fs_open(fsname, flags, 0, 0, unix_io_manager, &fs);
    if (retval)
    {
        printf("Could not open filesystem using %s\n", fsname);
        return 1;
    }

    retval = ext2fs_read_bitmaps(fs);
    if (retval)
    {
        printf("Could not read bitmaps from %s\n", fsname);
        return 1;
    }

    printf("Filesystem with blocksize %d\n", fs->blocksize);

    /* Create some files. */

    for (i = 1; i <= 200; i++)
    {
        sprintf(basename, "file%04d", i);

        retval = image_create_file(fs, EXT2_ROOT_INO, basename, 0644,
                                   1000, 1000, &ino);

        if (retval)
        {
            printf("Could not create file %s\n", basename);
            return 1;
        }

        if (ext2fs_file_open(fs, ino, EXT2_FILE_WRITE | EXT2_FILE_CREATE, &file))
        {
            printf("Could not write file %s\n", basename);
            return 1;
        }

        ext2fs_file_flush(file);
        ext2fs_file_close(file);
    }

    list_dir_data.fs = fs;
    list_dir_data.filename = NULL;
    list_dir_data.link_count = 0;

    image_list_dir(fs, "", image_list_dir_proc, &list_dir_data);
    image_list_dir(fs, "", image_list_dir_proc, &list_dir_data);

    /* Close the filesystem image. */

    retval = ext2fs_flush(fs);
    if (retval)
    {
        printf("Error flushing filesystem in %s\n", fsname);
        exitcode = 1;
    }

    retval = ext2fs_close(fs);
    if (retval)
    {
        printf("Error closing filesystem in %s\n", fsname);
        exitcode = 1;
    }

    ext2fs_free(fs);
    return exitcode;
}
